
local versionNumber = "023"
local version = "PolyPrinter post-processing v" .. versionNumber .. " copyright 2012-2018 PolyPrinter"
print( version )

-- TODO: test line suppressions with new changes to section ending!!!

local x = os.clock()
--[[
Copyright 2012, 2013, 2014, 2015, 2016, 2017 PolyPrinter
May be used to post-process files for printing on PolyPrinters.
All other rights reserved.
]]--

local debugging = false		-- set to true when trying to diagnose problems. Set View, Params
-- cmd line e.g. --duet "\\files\kinetric\3DParts\authored\Calibration\CalibSteps and Hole.duet nopost.gcode"

local doSectionMatchTracing = false -- debugging 

local replaceOriginal = not debugging
local isDuet = false
local absoluteExtrudes = true  -- Marlin uses absolute E positions, Duet uses Relative

local bDoJerkModification = true
local solidFillFlowTweak = 1.0 -- adjust the extrusion of the Solid Fill (only)
--local perimeterFlowTweak = .70 -- adjust outside loops
local perimeterFlowTweak = 1.0 -- 0 to eliminate outside loops
local sectionsEndWithBlankCommentLine = false;  -- used to be able to be handled that way
--[[ A bit tricky
;
; *** Every 2 Layers ***
; POST_PROC_SKIP_PERIMETER
M117 skipped
;
; Reset extruder pos
G92 E0
; BEGIN_LAYER_OBJECT z=3.25
]]--

-- Regular molds have mostly been done with every *N Layers in the Machine setting set to 4
--   so that 2 are printed , the 2 are skipped, making a robust pattern that is relatively coarse.
-- finer molds may need a smaller number, probably if numPerimsToSkip is also reduced to 1. Then use 2 or 3 for every *N layers.
-- but note that this doesn't have enough gaps for most purposes, when layers are thin. So don't use skip of 1!

POST_PROC_SKIP_PERIMETER_PATTERN = "^; POST_PROC_SKIP_PERIMETER" -- if we see this we will skip the perimeter for this layer
local skipPerimeterNextLayer = false    -- we get the indication just before the next layer starts
local skipPerimeterThisLayer = false    -- we manage this as each layer starts

-- NORMAL SETUP: 
local numPerimsToSkip = 2; -- 0 - do not skip ever,  1 - skip 1, 2 - skip 2
local skippedPerimeterCount = 0;

-- Special selective beginning of skipped perims to leave lower walls intact, if thin (1 loop)
--local ignoreSkippedPerimetersBelowZ = 17 -- .125 * 25.4 -- set to zero to skip all, to some number to wait until that layer
local ignoreSkippedPerimetersBelowZ = 0 -- set to zero to skip all, to some number to wait until that layer


-- be careful using this!
local MIN_SUPPORT_HEIGHT = .0 -- no support of any kind below this height - removes support touching bed, for underside lettering


-- Gathered Settings: - these are reasonable defaults but will be replaced if the parameters are spotted going by
local	maxFlowRate_MM3pSEC = 99999
local maxSparseSpeed_MMpSEC = 250   
local	extrusionWidth_MM = .35
local	layerThickness_MM = .25   
local maxLayerThickness_MM = 0.0 -- Zero if not seen. note that with Kiss 1.6 there can also be a different Maximum: max_layer_thickness_mm
local filamentDiameter_MM = 1.75

RESET_E_PATTERN     = "^G92 E0";
BEGIN_LAYER_PATTERN = "^; BEGIN_LAYER_OBJECT";
END_LAYER_PATTERN   = "^; END_LAYER_OBJECT"

-- there are a number of different path types that we need to recognize
PERIMETER_PATTERN   = "^; \'Perimeter";
LOOP_PATTERN        = "^; \'Loop";
CROWN_PATTERN       = "^; \'Crown Path";
OTHER_PATH          = "^; \'.+ Path";       -- any other path - Raft, Pillar, Prime Pillar

DESTRING_PATTERN    = "^; \'Destring";      -- this can apply to both the Suck, or a Wipe


SPARSEFILL_PATTERN  = "^; \'Sparse Infill";  -- hopefully the first layer above stacked sparse?
SUPPORT_ANY_PATTERN = "^; \'Support" -- any type of support

-- we don't use this one:
STACKED_SPARSEFILL_PATTERN = "^; \'Stacked Sparse Infill";  -- stacked sparse- below the real solid infill e.g. ; 'Stacked Sparse Infill Path', 5.1 [feed mm/s], 140.0 [head mm/s]


SOLIDFILL_PATTERN = "^; \'Solid";  -- e.g. ; 'Solid Path', 6.8 [feed mm/s], 187.5 [head mm/s]


MOVEXYZEF_PATTERN	   		= "^G1 X.+Y.+Z.+E.+F"  -- fast match of basic possibility
MOVEXYZEF_PARSE_PATTERN	    = "^G1 X([-%d%.]+)%s+Y([-%d%.]+)%s+Z([-%d%.]+)%s+E([-%d%.]+)%s+F([-%d%.]+)" -- suitable for a full parse

MOVE4_PATTERN	   		= "^G1 X.+Y.+Z.+E"  -- fast match of basic possibility of a 4-axis move pattern
MOVE4_PARSE_PATTERN	    = "^G1 X([-%d%.]+)%s+Y([-%d%.]+)%s+Z([-%d%.]+)%s+E([-%d%.]+)" -- suitable for a full parse

MOVEXYEF_PATTERN	   		= "^G1 X.+Y.+E.+F"  -- fast match of basic possibility
MOVEXYEF_PARSE_PATTERN	    = "^G1 X([-%d%.]+)%s+Y([-%d%.]+)%s+E([-%d%.]+)%s+F([-%d%.]+)" -- suitable for a full parse

MOVEXYE_PATTERN	   		= "^G1 X.+Y.+E"  -- fast match of basic possibility of a move pattern
MOVEXYE_PARSE_PATTERN	    = "^G1 X([-%d%.]+)%s+Y([-%d%.]+)%s+E([-%d%.]+)" -- suitable for a full parse

MOVE_EF_PATTERN	   		= "^G1 E.+F"  -- fast match of basic possibility
MOVE_EF_PARSE_PATTERN	    = "^G1 E([-%d%.]+)%s+F([-%d%.]+)" -- suitable for a full parse

MOVE_E_PATTERN			= "^G1 E"  -- fast match of basic possibility of an extruder move pattern
MOVE_E_PARSE_PATTERN	= "^G1 E([-%d%.]+)" -- suitable for a full parse

-- ; *** Slowing to match Min Layer Time (speed multiplier is 0.079412) ***
SLOWING_DOWN_PATTERN = "^; *** Slowing"  -- fast match of basic possibility
SLOWING_DOWN_PARSE_PATTERN	= "speed multiplier is ([-%d%.]+)" -- suitable for a full parse (Kisslicer up to 1.5)
SLOWING_DOWN_PARSE_PATTERN2	= "([-%d%.]+)" -- suitable for a full parse (Kisslicer Version 1.6+)


-- NOTE: ZeroBrane transforms triple * into a single in comments!
-- ; ******* Estimate 1.836691 [s], Slowing by 0.020x to 50.6 [s] *******
--; *** Estimate 7.549387 [s], Slowing (0.654x) to 10.0 [s] *** Kiss 1.6.3
--; *** Estimate all extruders at 7.5 [s], Slowing 0.645x to 10.0 [s] *** KISS 2.037

TIME_ESTIMATE_PATTERN = "^; *** Estimate "  -- fast match of basic possibility
TIME_ESTIMATE_PARSE_PATTERN = "([-%d%.]+).-Slowing.-([-%d%.]+)"
TIME_ESTIMATE_PARSE_PATTERN2	= ".*[-%d%.].-([-%d%.]+)" -- ignore texts, take second field (Kisslicer Version 1.6+) if wording changes
-- not needed? TIME_ESTIMATE_PARSE_PATTERN3	= ".*[-%d%.].-([-%d%.]+)" -- ignore texts, take second field (Kisslicer 2.037

--; Dwell to cool previous layer
--G4 P18054
SLOWING_DWELL_PATTERN = "^; Dwell"  -- fast match of basic possibility

local currentLayerThickness_mm = .25 -- just a default
local currentLayerZ_mm = 0;

-- last known axis positions
local prevX = 0
local prevY = 0
local prevZ = 0
local prevE = 0     -- this is ALWAYS ABSOLUTE
local prevF = 0

function clamp( num, lowLimit, hiLimit )
	num = math.min( num, hiLimit )
	return math.max( num, lowLimit )
end

function round(num, mult)
  return math.floor(num * mult + 0.5) / mult
end

function linearInterpolatePct( fpct, minVal, maxVal )
	if ( minVal == nil ) then return maxVal end
	if ( maxVal == nil ) then return minVal end
	return minVal + fpct * ( maxVal - minVal );
end

function linearInterpolatePct_Clamp( fpct, minVal, maxVal )
	fpct = math.min( fpct, 1.0 )
	fpct = math.max( fpct, 0 )
	return minVal + fpct * ( maxVal - minVal );
end

function round2dec(num)
  return math.floor(num * 10^2 + 0.5) / 10^2
end

local roundEdecimals = 5
local roundEmult = 10^(roundEdecimals or 0)
function roundE(num)
  return math.floor(num * roundEmult + 0.5) / roundEmult
end

function getNumericSettingValue( assignmentLine )
	local dummy;
	local pattern = "([%w ]+)=%s*([%d%.-]+)"
	local key, value
	dummy, dummy, key, value = assignmentLine:find( pattern )
	--print( "key " .. key .. " value " .. value )
	return tonumber( value );
end

-- parse a GCODE line of the form: M205 X05
-- return M, 205, X, 205 all separated out
function parseGCodeSetting1Value( gcodeLine )
	local dummy;
	local pattern = "(%a)(%d+)%s*(%a)([-%d%.-]+)"
	local gtype, codeNum, axis1, axis1val
	dummy, dummy, gtype, codeNum, axis1, axis1val = gcodeLine:find( pattern )
	--print( "key " .. gtype .. " value " .. codeNum )
	return gtype, tonumber( codeNum ), axis1, tonumber( axis1val )
end


-- parse a GCODE line of the form: M201 X3000 Y3000
-- return M, 201, X, 3000, Y, 3000 all separated out
function parseGCodeSetting2Value( gcodeLine )
	--print( "parseGCodeSetting2Value" .. gcodeLine )

	local dummy;
	local pattern = "(%a)(%d+)%s*(%a)([-%d%.-]+)%s*(%a)([-%d%.-]+)"
	local gtype, codeNum, axis1, axis1val, axis2, axis2val
	dummy, dummy, gtype, codeNum, axis1, axis1val, axis2, axis2val = gcodeLine:find( pattern )
	--print( "key " .. gtype .. " value " .. codeNum )
	return gtype, tonumber( codeNum ), axis1, tonumber( axis1val ), axis2, tonumber( axis2val )
end

-- should only be called when line is known to use G01 X... Y... Z... E...
function parseGCodeMove4( gcodeLine )
	--print( "parseGCodeMove4" .. gcodeLine )

	local dummy;
	local pattern = MOVE4_PARSE_PATTERN
	local axisX, axisY, axisZ, axisE
	dummy, dummy, axisX, axisY, axisZ, axisE = gcodeLine:find( pattern )
	--print( "X " .. axisX .. " Y " .. axisY .. " Z " .. axisZ .. " E " .. axisE )
	return tonumber( axisX ), tonumber( axisY ), tonumber( axisZ ), tonumber( axisE )
end


-- should only be called when line is known to use G01 X... Y... E...
function parseGCodeMoveXYE( gcodeLine )
	--print( "parseGCodeMoveXYE" .. gcodeLine )

	local dummy;
	local pattern = MOVEXYE_PARSE_PATTERN
	local axisX, axisY, axisE
	dummy, dummy, axisX, axisY, axisE = gcodeLine:find( pattern )
	--print( "X " .. axisX .. " Y " .. axisY .. " E " .. axisE )
  
  local absoluteE = tonumber( axisE )
  if ( not absoluteExtrudes ) then
    absoluteE = prevE + absoluteE     -- change it to an absolute value
  end

	return tonumber( axisX ), tonumber( axisY ), absoluteE
end

-- should only be called when line is known to use G01 X... Y... Z... E... F...
function parseGCodeMoveXYZEF( gcodeLine )
	--print( "parseGCodeMoveXYZEF" .. gcodeLine )
	local dummy, dummy2, axisX, axisY, axisZ, axisE, feedrate = gcodeLine:find( MOVEXYZEF_PARSE_PATTERN )
	--print( "Z", axisZ )
	--print( "X " .. axisX .. " Y " .. axisY .. " Z" .. axisZ .. " E " .. axisE .. " F " .. feedrate )
  
  local absoluteE = tonumber( axisE )
  if ( not absoluteExtrudes ) then
    absoluteE = prevE + absoluteE     -- change it to an absolute value
  end

	return tonumber( axisX ), tonumber( axisY ), tonumber( axisZ ), absoluteE, tonumber( feedrate )
end

-- should only be called when line is known to use G01 X... Y... E... F...
function parseGCodeMoveXYEF( gcodeLine )
	--print( "parseGCodeMoveXYEF" .. gcodeLine )
	local dummy, dummy2, axisX, axisY, axisE, feedrate = gcodeLine:find( MOVEXYEF_PARSE_PATTERN )
	--print( "X " .. axisX .. " Y " .. axisY .. " E " .. axisE .. " F " .. feedrate )
  
  local absoluteE = tonumber( axisE )
  if ( not absoluteExtrudes ) then
    absoluteE = prevE + absoluteE     -- change it to an absolute value
  end
  
	return tonumber( axisX ), tonumber( axisY ), absoluteE, tonumber( feedrate )
end

-- should only be called when line is known to use G01 X... Y... E... F...
function parseGCodeMoveEF( gcodeLine )
	--print( "parseGCodeMoveEF" .. gcodeLine )
	local dummy, dummy2, axisE, feedrate = gcodeLine:find( MOVE_EF_PARSE_PATTERN )
	--print( " E " .. axisE .. " F " .. feedrate )
  
  local absoluteE = tonumber( axisE )
  if ( not absoluteExtrudes ) then
    absoluteE = prevE + absoluteE     -- change it to an absolute value
  end
  
	return absoluteE, tonumber( feedrate )
end


-- should only be called when line is known to use G01 E...
function parseGCodeMoveE( gcodeLine )
	--print( "parseGCodeMoveE" .. gcodeLine )

	local dummy;
	local pattern = MOVE_E_PARSE_PATTERN
	local axisE
	dummy, dummy, axisE = gcodeLine:find( pattern )
	--print( " E " .. axisE )
  
  local absoluteE = tonumber( axisE )
  if ( not absoluteExtrudes ) then
    absoluteE = prevE + absoluteE     -- change it to an absolute value
  end

	return absoluteE
end

function parseSlowdownFactor( gcodeLine )
	local dummy
	local factor
	dummy, dummy, factor = gcodeLine:find( SLOWING_DOWN_PARSE_PATTERN ) 
  if ( factor == nil ) then
    dummy, dummy, factor = gcodeLine:find( SLOWING_DOWN_PARSE_PATTERN2)
  end
	--print( " slowdown ", factor, " Z ", prevZ )
	return tonumber( factor )
end

function parseTimeEstimate( gcodeLine )
	local dummy
	local factor
  local time 
	dummy, dummy, time, factor = gcodeLine:find( TIME_ESTIMATE_PARSE_PATTERN ) 
  if ( factor == nil ) then
    print("first parse failed")
    dummy, dummy, factor = gcodeLine:find( TIME_ESTIMATE_PARSE_PATTERN2 )
  end
	--print( " time " .. time .. " factor ", factor )
  if ( factor == nil ) then return nil end
	return tonumber( factor )
end

-- old style ; BEGIN_LAYER_OBJECT z=1.999
-- ; BEGIN_LAYER_OBJECT z=0.250 z_thickness=0.275
function parseLayerBeginHeightAndThickness( gcodeLine )
	local dummy;
  
	local pattern = ".*z=([%d%.]+).*z_thickness=([%d%.]+)" -- 
	local height
  local thickness_mm
	dummy, dummy, height, thickness_mm = gcodeLine:find( pattern )
  if ( height == nil ) then
    	dummy, dummy, height = gcodeLine:find( ".*z=([%d%.]+)" )  -- fallback to just height
  end
  if ( thickness_mm == nil ) then
    --print( "layer begin height value " .. height  )
    return tonumber( height ), nil
  else
    --print( "layer begin height value " .. height .. " thickness " .. thickness_mm )
    return tonumber( height ), tonumber( thickness_mm )
  end
end

-- parse a GCODE line of the form: M201 X3000 Y3000 Znnn where any of them may be missing or re-ordered
-- returns the words that make up the line
function parseGCodeWords( gcodeLine )

	local words = {}
	-- TODO: include all possible code letters
	--            only read up to comment indicator
	for word in line:gmatch( "[A-Z][-%d.]+" ) do table.insert( words, word ) end

	return words;
end


function makeGCode5Value( type, codeNum, axis1, axis1val, axis2, axis2val, axis3, axis3val, axis4, axis4val, axis5, axis5val )
	return table.concat( { type, codeNum, " " , axis1, axis1val, " ", axis2, axis2val, " ", axis3, axis3val, " ", axis4, axis4val, " ", axis5, axis5val } )
end

function makeGCode4Value( type, codeNum, axis1, axis1val, axis2, axis2val, axis3, axis3val, axis4, axis4val )
	return table.concat( { type, codeNum, " " , axis1, axis1val, " ", axis2, axis2val, " ", axis3, axis3val, " ", axis4, axis4val } )
end

function makeGCode3Value( type, codeNum, axis1, axis1val, axis2, axis2val, axis3, axis3val )
	--return type .. codeNum .. " " .. axis1 .. axis1val .. " " .. axis2 .. axis2val
	return table.concat( { type, codeNum, " " , axis1, axis1val, " ", axis2, axis2val, " ", axis3, axis3val } )
end

function makeGCode2Value( type, codeNum, axis1, axis1val, axis2, axis2val )
	--return type .. codeNum .. " " .. axis1 .. axis1val .. " " .. axis2 .. axis2val
	return table.concat( { type, codeNum, " " , axis1, axis1val, " ", axis2, axis2val } )
end

function makeGCode1Value( type, codeNum, axis1, axis1val )
	return type .. codeNum .. " " .. axis1 .. axis1val
end

function makeGCode0Value( axis1, axis1val )
	return axis1 .. axis1val
end

-- Vary the acceleration and jerk settings with quality setting
-- these can probably be extracted from the Prefix Gcode. It could have two
-- sets of settings, understood to encompass the range of quality settings

local DEFAULT_MAX_ACCEL = 3000
local DEFAULT_MIN_ACCEL = 500
local DEFAULT_MAX_JERK = 25
local DEFAULT_MIN_JERK =  5

-- default these to nil for easy management of the different possibilities
local maxAccelX = nil
local maxAccelY = nil

local minAccelX = nil
local minAccelY = nil

local maxJerkXY = nil
local minJerkXY = nil

local quality = 1
local foundAccelSpec = false
local foundJerkSpec = false
local inMainGCode = false
local didModifyAccel = false			-- will be true if the last settings were not the default


-- strings are unique in Lua, so they serve as unique, fast identifiers
-- acceleration handling states:
local stateNull = nil
local stateEmitAccelOnExtrudeMove = "stateEmitAccelOnExtrudeMove"
local state = stateNull

-- fan and slowdown handling states
local slowdownstate = nil
local slowdownState_FixUpFan = "slowdownState_FixUpFan"
local current_fan255 = 0; -- assume fan is off to begin with
-- TODO: parse all M106 commands coming through, so this is always current

-- NOT TESTED
local sectionState = nil
local stateSuppressOutput = "stateSuppressOutput"

local currentAccelX = maxAccelX
local currentAccelY = maxAccelY
local currentJerkXY = maxJerkXY   -- in mm/sec^2 (note that Duet uses (mm/min)/sec

-- accumulate time each layer, to give us a handle on layer times
local accumulatedTime_SEC = 0;
local accumulatedCautionaryTime_SEC = 0;

function clearAccumulatedTime()
  accumulatedTime_SEC = 0
  accumulatedCautionaryTime_SEC = 0
end

function getAccumulatedTime()
  return accumulatedTime_SEC
end

function getProbableActualFastestAccumulatedTime()
  return accumulatedCautionaryTime_SEC
end

-- keep our own current actual as-written command state variables
local prevOutX = 0  -- normally the same as prevX but we'll just keep our own
local prevOutY = 0
local prevOutZ = 0
local prevOutE = 0  -- prevE is the unmodified one, we get a modified one - This is ALWAYS ABSOLUTE.

local prevOutF = 0  -- this is the actual one sent, possibly modified. prevF is unmodified
local TIME_ESTIMATE_OPTIMISM_FACTOR = .5  -- how much we might be off, on our estimated time. ( actual time ) * this = our estimate
local cautionarySlowdownFactor = 1.0 -- if we detect an estimated layer time (for the previous layer) that looks too low, we can apply this factor to the speed. Never more than 1.0

-- thin layers should not need the same cooling time limit - it should allow faster speeds
local LAYER_THICKNESS_FOR_STANDARD_LAYER_TIME_MM = .25

function resetOutE()
  prevOutE = 0
  --print("reset out E")
end

function estimateTimeTo( ofile, g, x, y, z, e, f )
  -- this can be very simple, to begin with
  --ofile:write( "x ", x, ", y ", y, ", z ", z, ", e ", e, ", f ", f, "\n" )
-- to make an intentional error ofile = nil
  local xDist_mm = x - prevOutX
  local yDist_mm = y - prevOutY
  local zDist_mm = math.abs( z - prevOutZ )
  local eDist_mm = math.abs( e - prevOutE )
  
  local xyDist_mm = math.sqrt( xDist_mm * xDist_mm + yDist_mm * yDist_mm )
   
  --ofile:write( "xDist_mm ", xDist_mm, " yDist_mm ", yDist_mm, " xyDist_mm ", xyDist_mm, "\n" )
  --ofile:write( " zDist_mm ", zDist_mm, " eDist_mm ", eDist_mm, "\n"  )
  
  -- assume non-drawing moves are very fast
  if ( zDist_mm == 0 and eDist_mm == 0 ) then
    g = 0
  end
  
  local f_MMpSEC = f / 60
  local cautionaryFeedRate = f_MMpSEC * cautionarySlowdownFactor

   -- only one of these will really be the time that limits the move. It will be the longest one.
  -- that will virtually always be XY, unless XY is zero
  local limitingDistance_MM = math.max( xyDist_mm, zDist_mm, eDist_mm )

  
  if ( g == 0 ) then
    -- fast move for XY, roughly "f" for the others
    if ( xyDist_mm > 0 ) then 
      f_MMpSEC = 250  -- do not modify the cautionary one 
      cautionaryFeedRate = 250 -- same if cautionary
    end
  end
  

  --ofile:write ( "limitingDistance_MM ", limitingDistance_MM )
  local fastestPossibleTime_SEC = limitingDistance_MM / f_MMpSEC
  local fastestPossibleCautionaryTime_SEC = limitingDistance_MM / cautionaryFeedRate
  --ofile:write ( "f_MMpSEC ", f_MMpSEC )
  --ofile:write ( " fastestPossibleTime_SEC ", fastestPossibleTime_SEC )
  --ofile:write ( " cautionaryFeedRate ", cautionaryFeedRate )
  --ofile:write ( " fastestPossibleCautionaryTime_SEC ", fastestPossibleCautionaryTime_SEC, "\n" )
  
  prevOutX = x
  prevOutY = y
  prevOutZ = z
	prevOutE = e  -- prevE is the unmodified one, we get a modified one (always Absolute)
  prevOutF = f  -- this is the actual one intended to be sent, possibly modified (it will be written out differently than this, if cautionarySlowdown is happening)

  return fastestPossibleTime_SEC, fastestPossibleCautionaryTime_SEC
end

-- this takes only the "normal" time estimate, before the cautionary slowdown has been figured into it
-- side effect - changes prevOutE to current value
function accumulateTimeTo( ofile, x, y, z, e, f )
  local time, ctime = estimateTimeTo( ofile, 1, x, y, z, e, f )
  -- since this is called everywhere, we may as well use it to keep track of what's been happening
  --ofile:write( "; will take " .. time .. "\n" )
  accumulatedTime_SEC = accumulatedTime_SEC + time
  --ofile:write ( " adding ", ctime, "\n" )
  accumulatedCautionaryTime_SEC = accumulatedCautionaryTime_SEC + ctime
  --ofile:write ( " accumulatedCautionaryTime_SEC ", accumulatedCautionaryTime_SEC, "\n" )
 return time
end

function writeCurrentAccelerationLine( ofile )
	ofile:write( makeGCode2Value( "M", 201, "X", currentAccelX, "Y", currentAccelY ) , "\n" )
end

function writeCurrentJerkLine( ofile )
	if ( bDoJerkModification ) then
    if ( isDuet ) then
      ofile:write( makeGCode2Value( "M", 566, "X", currentJerkXY * 60, "Y", currentJerkXY * 60 ), "\n" )
    else
      ofile:write( makeGCode1Value( "M", 205, "X", currentJerkXY ), "\n" )
   end
	end
end

function writeGivenELine( ofile, absE )
  local e = absE
  if ( not absoluteExtrudes ) then
    -- convert to relative extrusion
     e = e - prevOutE -- make e relative to prev absolute value
  end
	ofile:write( makeGCode0Value( "G1 E", roundE(e) ) , "\n" )
  accumulateTimeTo( ofile, prevOutX, prevOutY, prevOutZ, absE, prevOutF )
end

function writeGivenZFLine( ofile, z, f )
	ofile:write( makeGCode1Value( "G1 Z", z, "F", math.floor( f * cautionarySlowdownFactor ) ) , "\n" )
  accumulateTimeTo( ofile, prevOutX, prevOutY, z, prevOutE, f )
end

function writeGivenEFLine( ofile, absE, f )
  local e = absE
  if ( not absoluteExtrudes ) then
    -- convert to relative extrusion
    e = e - prevOutE -- make e relative to prev absolute value
  end
	ofile:write( makeGCode1Value( "G1 E", roundE(e), "F", math.floor( f * cautionarySlowdownFactor ) ) , "\n" )
  accumulateTimeTo( ofile, prevOutX, prevOutY, prevOutZ, absE, f )
end

function writeGivenXYELine( ofile, x, y, absE )
  local e = absE
  if ( not absoluteExtrudes ) then
    -- convert to relative extrusion
    e = e - prevOutE -- make e relative to prev absolute value
  end
	ofile:write( makeGCode3Value( "G1 X", x, "Y", y, "E", roundE(e) ) , "\n" )
  accumulateTimeTo( ofile, x, y, prevOutZ, absE, prevOutF )
end

function writeGivenXYEFLine( ofile, x, y, absE, f )
  local e = absE
  if ( not absoluteExtrudes ) then
    -- convert to relative extrusion
    e = e - prevOutE -- make e relative to prev absolute value
  end
	ofile:write( makeGCode4Value( "G", 1, "X", x, "Y", y, "E", roundE(e), "F", math.floor( f * cautionarySlowdownFactor ) ) , "\n" )
  accumulateTimeTo( ofile, x, y, prevOutZ, absE, f )
end

function writeGivenXYZELine( ofile, x, y, z, absE )
  local e = absE
  if ( not absoluteExtrudes ) then
    -- convert to relative extrusion
    e = e - prevOutE -- make e relative to prev absolute value
  end
	ofile:write( makeGCode4Value( "G", 1, "X", x, "Y", y, "Z", z, "E", roundE(e) ) , "\n" )
  accumulateTimeTo( ofile, x, y, z, absE, prevOutF )
end

function writeGivenXYZEFLine( ofile, x, y, z, absE, f )
  local e = absE
  if ( not absoluteExtrudes ) then
    -- convert to relative extrusion
    e = e - prevOutE -- make e relative to prev absolute value
  end
	ofile:write( makeGCode5Value( "G", 1, "X", x, "Y", y, "Z", z, "E", roundE(e), "F", math.floor( f * cautionarySlowdownFactor ) ) , "\n" )
  accumulateTimeTo( ofile, x, y, z, absE, f )    -- accumulate without taking into account the cautionary slowdown we will apply
end

function writeFanControlLine( ofile, fan )
  --print("fan"..fan.. "\n");
	ofile:write( makeGCode1Value( "M", 106, "S", math.floor( fan ) ), "\n" )
end


--[[
There are lots of different sections in the Kss slicer output:
; 'Skirt', 4.4 [feed mm/s], 60.0 [head mm/s]
; 'Wipe (and De-string)', 0.0 [feed mm/s], 60.0 [head mm/s]
; 'Solid', 2.2 [feed mm/s], 60.0 [head mm/s]
; 'Stacked Sparse Infill', 6.3 [feed mm/s], 151.7 [head mm/s]

The most important one for us to modify is:
; 'Perimeter', 2.2 [feed mm/s], 60.0 [head mm/s]

Fairly important:
; 'Loop', 2.2 [feed mm/s], 60.0 [head mm/s]
; 'Crown', 2.6 [feed mm/s], 91.2 [head mm/s]


Sections appear to end with a line with a single ';' so we can probably use that
to restore the default settings at the end of the sections we modify.
Alternatively, we set the accelerations every time any section begins.
]]--
local lastPctFast = -1

function writeAccelSettings( ofile, pctFast )
	pctFast = clamp( pctFast, 0, 1.0 )
	if ( pctFast ~= lastPctFast ) then
		-- write out a line that sets it, based on the quality setting
		currentAccelX = math.ceil( linearInterpolatePct( pctFast, minAccelX, maxAccelX ) )
		currentAccelY = math.ceil( linearInterpolatePct( pctFast, minAccelY, maxAccelY ) )
		currentJerkXY = math.ceil( linearInterpolatePct( pctFast, minJerkXY, maxJerkXY ) )

		-- acceleration and jerk sttings take place immediately, no buffered like Gcode.
		-- simplest is to drain the buffer first.
		-- NOTE: this causes a stall doing the last action because the planner always executes it at slow speed, hoping for more
		--             - set a flag, or drain to 1 left, or handle accel changes in synch (pick up when executed)
		-- not good enough by itself fout:write( "M400\n" )

		writeCurrentAccelerationLine( ofile )
		writeCurrentJerkLine( ofile )
		didModifyAccel = ( pctFast < 1.0 )
		lastPctFast = pctFast
	end
	--if ( pctFast >= 1.0 ) then
	--	didModifyAccel = false  -- back to defaults
	--end
end

function AfterInit()
--print( "maxAccelX", maxAccelX, "minAccelX", minAccelX )
--print( "maxAccelY", maxAccelY, "minAccelY", minAccelY )
--print( "maxJerkXY", maxJerkXY, "minJerkXY", minJerkXY )	-- fix up acceleration and jerk, depending on the state
	maxAccelX = maxAccelX or DEFAULT_MAX_ACCEL
	minAccelX = minAccelX or DEFAULT_MIN_ACCEL
	maxAccelY = maxAccelX or DEFAULT_MAX_ACCEL
	minAccelY = minAccelX or DEFAULT_MIN_ACCEL
	maxJerkXY = maxJerkXY or DEFAULT_MAX_JERK
	minJerkXY = minJerkXY or DEFAULT_MIN_JERK


	if ( minJerkXY >= maxJerkXY ) then
		minJerkXY = DEFAULT_MIN_JERK
	end

	if ( minAccelX >= maxAccelX ) then
		minAccelX =  DEFAULT_MIN_ACCEL
	end

	if ( minAccelY >= maxAccelY ) then
		minAccelY = DEFAULT_MIN_ACCEL
	end

--print( "maxAccelX", maxAccelX, "minAccelX", minAccelX )
--print( "maxAccelY", maxAccelY, "minAccelY", minAccelY )
--print( "maxJerkXY", maxJerkXY, "minJerkXY", minJerkXY )
end

local kissMinLayerTime = 40           -- Kisslicer setting
local kissFanLoopPercent = 0
local kissFanInsidePercent = 0
local kissFanCoolPercent = 0        -- not used yet

local ourMinLayerTime = 10 * TIME_ESTIMATE_OPTIMISM_FACTOR     -- setting
local maxAllowedFan = 1.0
local minUseableFan = .10

-- the old way, before Kisslicer's own linear fan/layer time implementation, we had to artificially trigger Kisslicer 
-- to get involved, so we set layer time to 40 sec in the settings, and then used .25 to fix it up to 10 sec.
-- with Kisslicer version 1.6 and higher, we'll start using an actual 10-sec setting.
-- but we really want to keep the postproc compatible with Kisslicer 1.5 for a while
-- so we will change this value when running the newer versions.
local SLOWDOWN_FOR_FULL_FAN = .25 -- if slowdown hits this factor, use full fan.
local doModifyLayerTimes = true   -- if ture, we are leaving all speeds and fan settings alone, hoping Kiss 1.6 handles it all correctly like we want.

-- Our own prefix script has a line like: M201 X3000 Y3000
-- the easiest thing is to substitute it with numbers based on the quality settings
-- but if it's not found, we need to insert one.

--local didSparseThisLayer = false;    -- cleared at the beginning of every layer
--local currentLayerHeight = 0;		-- tracked as layers go by
local slowdownFactor = 1.0;       -- normally 1, otherwise is set by Kisslicer according to initial calculations
local speedupFactor  = 1.0;        -- this is how much faster (or slower) than the original gcode we'll try to go (limited by flow rate) when starting to apply fan instead of slowing down

local currentFlowTweak = 1.0
local sectionEInputBase = 0				-- the base input E coming in to the section
local sectionEOutputBase = 0				-- the base E for output
local currEout = 0					-- calculated current E based on flow tweak and input E
local layerCount = 0      -- 0 until official beginning of a layer is seen
local maxDrawSpeedForFlowRateLimit_MMpMIN = 9999999 -- safe default. This gets calculated when main G-code starts
local prevLayerTime_SEC = 0;
local totalPrintTime_SEC = 0;

-- returns the lowest fan speed we should use. Set by the lower of the two Kisslicer material settings.
function minimumFanPercent()
  return math.min( kissFanInsidePercent, kissFanLoopPercent )
end

function resetAllTweaks( fout )
	-- restore the default (max) settings
  if ( didModifyAccel ) then
    writeAccelSettings( fout, 1.0 )  -- full speed settings
  end
  
  if ( currentFlowTweak ~= 1.0 ) then
    currentFlowTweak = 1.0	-- reset flow tweak after each section
    -- set up for next section extrusion
    sectionEInputBase = prevE
    sectionEOutputBase = currEout
  end
 
  sectionState = nil      -- end supressing output 
end

local currAccelPctFast = 1

function writeCurrAccelSettings( ofile )
  -- ease up on the acceleration when slowing down for cooling. This gives a more proportional effect.
	writeAccelSettings( ofile, currAccelPctFast * slowdownFactor * cautionarySlowdownFactor );
end



-- ************************ Start ****************************
-- TESTING

if ( arg[1] == "--duet" ) then
  isDuet = true
  absoluteExtrudes = false
  arg[1] = arg[2]
  print("for Duet board")
end

print( arg[1] )
local inFileName  = arg[1]
if ( inFileName == null ) then
	replaceOriginal = false
	inFileName = "extrudetest.gcode"
	print( "TESTING ONLY - no input name found - using " .. inFileName )
end

-- due to a Windows file path length limitation in the older API, Lua has trouble opening files with a total path length > 260 characters
-- So, we could skirt this somewhat (and possibly break the Mac install) by modifying the name to make Windows/Lua at least not barf on it.
--local longname = "\\\\?\\UNC" .. string.gsub( inFileName, "\\\\", "\\")
--print("longname " .. longname)
--inFileName = longname

-- we cannot assume that lengthening the name is acceptable 
-- so replace some of it without lengthening it
local outFileName = string.gsub(inFileName, ".gcode", versionNumber .. ".g" )
local fin  = assert( io.open( inFileName ) )

--local fout = assert( io.open( outFileName, "wb" ) )  -- binary mode saves a few characters - but it's not easily editable by Notepad then
local fout = assert( io.open( outFileName, "w" ) )
--print("opened output " .. outFileName)

fout:setvbuf("full", 2^24)
fout:write( "; ", version , '\n' )

-- Flow Tweaking: incoming E values are what we always use for making decisions about parsing the file
--         only the output is modified according to the flow tweak.
local skipNextLine = false

for line in fin:lines() do
	local doEcho = true  -- by default we always write out the original line at the bottom of this loop, but it can be suppressed
	--fout:write( line .. "\n" );
	--print( line  );
  if ( sectionState == stateSuppressOutput or skipNextLine ) then 
    --print("suppressing")
    if ( skipNextLine ) then skipNextLine = false end
    doEcho = false;   -- no output desired
    if ( sectionsEndWithBlankCommentLine ) then
      -- but watch for section end
      if ( line:match( "^;$" ) ) then  -- single comment line - end of section
        resetAllTweaks( fout )
      end
    elseif ( line:match( OTHER_PATH ) and not line:match(DESTRING_PATTERN) ) then    
      -- while suppressing, any Path type at all is probably a reasonable place to stop suppressing???
      resetAllTweaks( fout )
    end
	elseif ( line:match("^G1" ) ) then
		if ( line:match( MOVEXYZEF_PATTERN ) ) then
      --print( "MOVEXYZEF_PATTERN" )
			-- parse it and remember the axis positions
			local x, y, z, e, f = parseGCodeMoveXYZEF( line )
      --print( z, e, f )
			if ( e == nil ) then
				print( "MOVEXYZEF_PATTERN e is nil " .. line .. " prevE " .. prevE )
			end
      if ( f == nil ) then
				print( "MOVEXYZEF_PATTERN f is nil " .. line .. " prevF " .. prevF )
			end
	
			currEout = sectionEOutputBase + currentFlowTweak * ( e - sectionEInputBase )
      if ( currentFlowTweak == 0 ) then
	      -- write this one even if currentFlowTweak is 0, because we don't want to miss the Z
        -- we don't want to move XY in case it drools where we want empty space.
        writeGivenZFLine( fout, z, f )  -- the F will probably be wrong. Firmware will clamp OK.
        doEcho = false
        
      elseif ( currEout ~= e ) then
				-- the e value is modified - substitute our own output
				doEcho = false
				-- write out the modified line

				writeGivenXYZEFLine( fout, x, y, z, currEout, f );
			else
				-- echo the line - the E value is unchanged
        accumulateTimeTo( fout, x, y, z, e, f )
			end
      
      -- save original command values
      prevX = x
			prevY = y
			prevZ = z
			prevE = e
      prevF = f

		elseif ( line:match( MOVE4_PATTERN ) ) then
			-- parse it and remember the axis positions
			local x, y, z, e = parseGCodeMove4( line )
			if ( e == nil ) then
				print( "MOVE4_PATTERN e is nil " .. line .. " prevE " .. prevE )
			end
	
			currEout = sectionEOutputBase + currentFlowTweak * ( e - sectionEInputBase )
			if ( currEout ~= e ) then
				-- the e value is modified - substitute our own output
				doEcho = false
				-- write out the modified line

				if ( currentFlowTweak > 0 ) then 
          writeGivenXYZELine( fout, x, y, z, currEout ) 
        else
          writeGivenXYZLine( fout, x, y, z ) 
        end
			else
				-- echo the line - the E value is unchanged
	      accumulateTimeTo( fout, x, y, z, e, prevF )
      end
      
      -- as originally commanded
      prevX = x
			prevY = y
			prevZ = z
			prevE = e

  elseif ( line:match( MOVEXYEF_PATTERN ) ) then
      --fout:write("MOVEXYEF_PATTERN\n")
      -- this typically occurs at the begining of each extrusion section i.e. perim, infill, etc.
			-- parse it and remember the axis positions, and grab the feedrate
			local x, y, e, f = parseGCodeMoveXYEF( line )

			local eDiff = e - prevE

			if ( eDiff > 0 ) then
				-- it's an extruding move with feedrate set

				-- also write the changed acceleration if that's appropriate
				if ( state == stateEmitAccelOnExtrudeMove ) then
					writeCurrAccelSettings( fout );
					state = stateNull
				end

				if ( speedupFactor ~= 1 ) then
          -- we're actually going to print faster or slowe than the original gcode says to.
					-- modify the feed rate back up to full speed
					f = f * speedupFactor
          
          -- in case it is > 1, also apply the max flow rate limit, because Kisslicer does not document that limit's effect separately
          -- and gives us a "Fan" slowdown percentage that is based on the layer time BEFORE flow rate limits are applied
          f =  math.min( f, maxDrawSpeedForFlowRateLimit_MMpMIN )
					f =  math.ceil( f )
          
					doEcho = false;     -- absorb the line because we need to modify it
        end
       
  
			end
			currEout = sectionEOutputBase + currentFlowTweak * ( e - sectionEInputBase )
			if ( currEout ~= e or doEcho == false ) then
				-- the e value is modified - substitute our own output
				doEcho = false
				-- write out the modified line

				if ( currentFlowTweak > 0 ) then writeGivenXYEFLine( fout, x, y, currEout, f ) end
			else
				-- echo the line - the E value is unchanged
        accumulateTimeTo( fout, x, y, prevZ, e, f )
			end
      
      -- a bit of housekeeping, because this type of line typically begins a new layer, but after the SLOWDOWN line if it exists.
             -- if there is no speedup factor anymore, this layer does not use the fan
      if ( slowdownState ~= slowdownState_FixUpFan and current_fan255 > 0 ) then   -- this test ensures we only check this once, after a layer that used the fan, since it's only set when the kisslicer slowdown line is seen
        -- we had previously (previous layer) been using the fan (probably)
        local fan255 = minimumFanPercent() * 255
        if ( current_fan255 ~= fan255 ) then
          writeFanControlLine( fout, current_fan255 ); -- turn off the fan or at least use the smallest setting
        end
      end
      
      prevX = x
			prevY = y
			prevE = e
      prevF = f
  
		elseif ( line:match( MOVEXYE_PATTERN ) ) then
			-- parse it and remember the axis positions
			local x, y, e = parseGCodeMoveXYE( line )

			if ( e == nil ) then
				print( "MOVEXYE_PATTERN e is nil " .. line )
			end

			if ( state == stateEmitAccelOnExtrudeMove ) then
				--local xDiff = x - prevX
				--local yDiff = y - prevY
				local eDiff = e - prevE
				-- if we see a real extrusion move, first emit the latest modified aceleration settings
				-- do the real parse of the move, and see if it there is extrusion in the move.
				-- for this we must have parsed ALL G1 lines and kept track of the E position
				-- it could be a retract, but not likely. However we are looking for a positive extrusion anyway.
				-- It might include the Prime movement.
				if ( eDiff > 0 ) then
					-- it's an extrusion. We really don't need to check to see if there's an X,Y move too
					-- for it to be the point where we need to set the current accelerations
					writeCurrAccelSettings( fout );
					state = stateNull
				end

			end
			--print( "MOVE4_PATTERN e" .. e )
			currEout = sectionEOutputBase + currentFlowTweak * ( e - sectionEInputBase )
			if ( currEout ~= e ) then
				-- the e value is modified - substitute our own output
				doEcho = false
				-- write out the modified line

				if ( currentFlowTweak > 0 ) then writeGivenXYELine( fout, x, y, currEout ) end
			else
				-- echo the line - the E value is unchanged
        accumulateTimeTo( fout, x, y, prevZ, e, prevF )
			end
      
			prevX = x
			prevY = y
			prevE = e


		elseif ( line:match( MOVE_EF_PATTERN ) ) then
			-- parse it and remember the axis position
			e, f = parseGCodeMoveEF( line )
			if ( e == nil ) then
				print( "MOVE_EF_PATTERN e is nil" )
			end

			currEout = sectionEOutputBase + currentFlowTweak * ( e - sectionEInputBase )
			if ( currEout ~= e ) then
				-- the e value is modified - substitute our own output
				doEcho = false
				-- write out the modified line

				if ( currentFlowTweak > 0 ) then writeGivenEFLine( fout, currEout, f ) end
			else
				-- echo the line - the E value is unchanged
        accumulateTimeTo( fout, prevX, prevY, prevZ, e, f )
			end
      prevE = e
      prevF = f
      
		elseif ( line:match( MOVE_E_PATTERN ) ) then
			-- parse it and remember the axis position
			e = parseGCodeMoveE( line )
			if ( prevE == nil ) then
				print( "MOVE_E_PATTERN e is nil" )
			end

			currEout = sectionEOutputBase + currentFlowTweak * ( e - sectionEInputBase )
			if ( currEout ~= e ) then
				-- the e value is modified - substitute our own output
				doEcho = false
				-- write out the modified line

				if ( currentFlowTweak > 0 ) then writeGivenELine( fout, currEout ) end
			else
				-- echo the line - the E value is unchanged
        accumulateTimeTo( fout, prevX, prevY, prevZ, e, prevF )
			end
      prevE = e
		end

	elseif ( line:match( PERIMETER_PATTERN ) ) then

		if doSectionMatchTracing then print "found Perimeter" end

		local MAX_PERIM_PctFast = .5  -- clamp it for the perimeter - always limit the acceleration a bit
		local pctFast = 1.0 - quality
		pctFast = math.min( pctFast, MAX_PERIM_PctFast );
		currAccelPctFast = pctFast  -- save for when first real extrude move is found
		state = stateEmitAccelOnExtrudeMove
    

    if ( skipPerimeterThisLayer ) then
      currentFlowTweak = 0;
    else
      currentFlowTweak = perimeterFlowTweak
    end
    
		sectionEInputBase = prevE
		sectionEOutputBase = currEout

  elseif ( ( prevZ >= ignoreSkippedPerimetersBelowZ ) 
        and line:match( POST_PROC_SKIP_PERIMETER_PATTERN ) ) then
    -- this pattern occurs between layers
		-- we will skip the perimiter for the next layer
    skipPerimeterNextLayer = true
    print("skipping next")

  -- TODO: remove this skip when Kisslicer has a good and necessary Dwell
  --elseif ( line:match( SLOWING_DWELL_PATTERN ) ) then
    -- eat the G4 line
  --  skipNextLine = true
	elseif ( line:match(TIME_ESTIMATE_PATTERN) or line:match( SLOWING_DOWN_PATTERN ) ) then
--print( line )
    if ( doModifyLayerTimes ) then
      local kissSlowdownFactor = parseTimeEstimate( line )  -- this is what Kiss has calculated. We override it.
      --print( "kissSlowdownFactor " .. kissSlowdownFactor )
      if ( kissSlowdownFactor == nil ) then
        kissSlowdownFactor = parseSlowdownFactor( line )  
      end
      
      -- we always modify the fan settings when Kisslicer messes with slowdown
      slowdownState = slowdownState_FixUpFan -- if we see a move with a feedrate... adjust it and make sure the fan is used instead up to a point
      
       -- we may modify the feedrate by feeding in more fan, up to a point
      local fanfraction = 0;  -- default to no fan
      if ( kissSlowdownFactor >= SLOWDOWN_FOR_FULL_FAN ) then
        fanfraction = 1 - ( kissSlowdownFactor - SLOWDOWN_FOR_FULL_FAN ) / ( 1 - SLOWDOWN_FOR_FULL_FAN )
        -- modify the feed rate back up to full speed
        speedupFactor = 1 / kissSlowdownFactor
        slowdownFactor = 1 -- not actually slowing down, either, unless Kisslicer is wrong. This is only used for acceleration modification.
       --print( "Fan", fanfraction, "slowdown", slowdownFactor )
      else
        -- we're at the slowdown lower limit - below this the fan has to be full on,
        -- and we will keep a slowdown factor so that the time is a fixed fraction of the
        -- set minimum layer time
        fanfraction = 1
        -- modify the speed based on a new effective slowdown factor as compared to the minimum
        -- so at minimum it's full speed, and below that it's proportional to the feed factor requested
        -- e.g. at 1/2 the slowdown for full fan it is e.g. .25 requested, now becomes .5 because it's .5 of /25
        -- the speedupFactor represents how much faster than the original gcode we will go. This will still be slower than if no slowdown was happening.
        speedupFactor = ( 1 / SLOWDOWN_FOR_FULL_FAN )     -- this results in a fixed speed multiplier e.g. 4x
        slowdownFactor = kissSlowdownFactor * speedupFactor  -- e.g. if Kiss says .1, and we do 4x, then this is .4, or 40% of the nominal layer speed parameters.
         -- NOTE: in the actual extrusion line that sets the feed factor F, we wil be applying a further modification
        --       that limits the flow rate to the specified limit for the material. This may reduce the actual feed rate, and effective slowdown factor
        --print( "Fan", fanfraction, "slowdown", slowdownFactor )

      end
      
      -- set the fan to the required speed (or a minimal speed so it's not stalled)
      local fan255 = math.max( minimumFanPercent(), linearInterpolatePct( fanfraction, minUseableFan, maxAllowedFan ) ) * 255
      if ( fan255 ~= current_fan255 ) then
        writeFanControlLine( fout, math.ceil( fan255 ) );
        current_fan255 = fan255
      end
    end -- if doModifyLayerTimes
    
	elseif ( line:match( LOOP_PATTERN ) or line:match( CROWN_PATTERN ) ) then

		if doSectionMatchTracing then print "found Loop" end

		-- a heuristic: half way to the quality setting?
		local pctFast = 1.0 - quality * .5  -- average between full speed and the quality setting speed

		currAccelPctFast = pctFast  -- save for when first real extrude move is found
		state = stateEmitAccelOnExtrudeMove

	elseif ( line:match( SOLIDFILL_PATTERN ) ) then

		if doSectionMatchTracing then print "found Solid Fill" end
    resetAllTweaks( fout )
		currentFlowTweak = solidFillFlowTweak
		sectionEInputBase = prevE
		sectionEOutputBase = currEout

	elseif ( line:match( SUPPORT_ANY_PATTERN ) ) then

		if doSectionMatchTracing then print "found Support" end
    resetAllTweaks( fout )

    if ( prevZ < MIN_SUPPORT_HEIGHT ) then
      -- stop printing until end of section
      currentFlowTweak = 0               -- don't allow extruder feeds to accumulate
      
    end

--[[ -- not needed
	elseif ( line:match( SUPPORT_INTERFACE_PATTERN ) ) then

		if doSectionMatchTracing then print "found Support Interface" end

    if ( prevZ > MIN_SUPPORT_INTERFACE_WEAKENING_HEIGHT ) then
      currentFlowTweak = supportInterfaceFlowTweak
      sectionEInputBase = prevE
      sectionEOutputBase = currEout
    end
]]--
  elseif ( line:match(DESTRING_PATTERN) ) then
    -- we will ignore destrings for now - but this keep it from triggering an end-of-section below

	elseif ( sectionsEndWithBlankCommentLine and line:match( "^;$" ) ) then  -- single comment line - end of section
  -- NOTE: it's possible we'd be better off some day to notice ALL section types beginning rather than count on this match
		if doSectionMatchTracing then print "found end of section" end
    resetAllTweaks( fout )
		--doEcho = false
	elseif ( line:match( OTHER_PATH ) ) then  -- end of path, and not recognized, so it is an end-of-section 
		if doSectionMatchTracing then print "found end of section" end
    resetAllTweaks( fout )
		--doEcho = false    

	elseif ( line:match( RESET_E_PATTERN ) ) then
		if doSectionMatchTracing then print "found reset E" end
    -- actually, we should really read the arg!
    resetOutE()
  
	elseif ( line:match( BEGIN_LAYER_PATTERN ) ) then -- ; BEGIN_LAYER_OBJECT z=1.999
		if doSectionMatchTracing then print "found Layer Begin" end
  -- print( "Prev layer (at", prevZ, ") would have taken ", round2dec( prevLayerTime_SEC ) )
		slowdownFactor = 1.0 -- reset every layer in case it's not there
    speedupFactor = 1.0
    slowdownState = nil -- wait and see if we find a slowdown pattern
    local newLayerZ, newLayerThickness = parseLayerBeginHeightAndThickness( line )
    if ( newLayerThickness == nil ) then
      currentLayerThickness_mm = newLayerZ - currentLayerZ_mm
      -- ALSO: Kiss 1.6 can have different Z heights for different parts on the same bed! In that case, this
      --       is going to potentially be quite far off when using the Begin Z height for a "layer" since the prev layer
      --       might actually have been for a different part. Also, the elapsed time may be for the previous part!!!!
      currentLayerThickness_mm = clamp( currentLayerThickness_mm, layerThickness_MM, maxLayerThickness_MM ) -- at least make sure it's in a sensible range
    else
      currentLayerThickness_mm = newLayerThickness
    end
    
    currentLayerZ_mm = newLayerZ

    --print("new Z " .. newLayerZ .. " thickness " .. currentLayerThickness_mm )
   
		-- this is probably close to a G92 E0 so reset the extruder position (G92 is not emitted for relative E so watch out - this has to be done correctly!)
		prevE = 0
		currEout = 0
		currentFlowTweak = 1.0
    -- this section deals with slowing down if necessary. It is always reacting to the previous layer, so if 
    -- there are very large fundamental differences between adjacent layers it may react too late
    prevLayerTime_SEC = getAccumulatedTime()
    local newCautionarySlowdownFactor = 1.0
    
    -- if we are doing .10 now and the standard is .20, then the layer time can be half the nominal one
    local layerTimeThicknessFactor = currentLayerThickness_mm / LAYER_THICKNESS_FOR_STANDARD_LAYER_TIME_MM
    local minLayerTimeForThickness_sec = ourMinLayerTime * layerTimeThicknessFactor
    --print("layerTimeThicknessFactor " .. layerTimeThicknessFactor .. " minLayerTimeForThickness_sec " .. minLayerTimeForThickness_sec)
  
    -- NOTE: this code is invoked only as a watchdog on top of all other calculations like using the fan to restore speeds
    --        that Kisslicer slowed down.
    if ( prevLayerTime_SEC < minLayerTimeForThickness_sec and prevZ >= .4 ) then -- only apply this after we're definitely clear of the bed. The bed will cool the pastic down very rapidly.
      newCautionarySlowdownFactor = prevLayerTime_SEC / minLayerTimeForThickness_sec
      cautionarySlowdownFactor = ( cautionarySlowdownFactor + newCautionarySlowdownFactor ) / 2 -- filters the result a lot
      --print("Warning: Layer at " .. prevZ .. " may be too fast- cautionary slowdown: " .. cautionarySlowdownFactor ) 
      print( "Prev layer (at", prevZ, ") would have taken ", round2dec( prevLayerTime_SEC ), " but modifed to take at least " .. round2dec( getProbableActualFastestAccumulatedTime() ) )
      fout:write( "; Prev layer would have taken ", round2dec( prevLayerTime_SEC ), " but modifed to take at least " .. round2dec( getProbableActualFastestAccumulatedTime() ) .. "\n" )
      --fout:write("; Warning: Layer at " .. prevZ .. " may be too fast - cautionary slowdown: " .. cautionarySlowdownFactor .. "\n" )
    end
    
    totalPrintTime_SEC = totalPrintTime_SEC + getProbableActualFastestAccumulatedTime() -- for the overall print time, use the one we actually wrote gcode for
    clearAccumulatedTime()
    resetOutE()
    
    if ( skipPerimeterNextLayer ) then
      -- we have seen the indicator for a layer to be skipped. Now act upon it.
      skipPerimeterThisLayer = true
      skipPerimeterNextLayer = false
      skippedPerimeterCount = 0 -- we will be skipping
      --print("will skip, begin sequence")
    end
    if ( skipPerimeterThisLayer ) then
      skippedPerimeterCount = skippedPerimeterCount + 1
      print("skip")
    end
    
		-- these two may not be required
		sectionEInputBase = prevE
		sectionEOutputBase = currEout
    layerCount = layerCount + 1 -- first found layer is 1
    if ( debugging ) then
      print( layerCount )
    end

elseif ( line:match( END_LAYER_PATTERN ) ) then
    if ( skippedPerimeterCount >= numPerimsToSkip ) then
      skipPerimeterThisLayer = false -- be sure not to keep skipping the perimeter
      skippedPerimeterCount = 0 -- just for cleanliness
      --print("end of skips")
    end
    
 		if doSectionMatchTracing then print "found Layer End" end

	elseif ( line:match( "Speed vs Quality" ) ) then

		--print "found quality setting"
		-- parse the setting
		quality = getNumericSettingValue( line )
		print( "Quality " .. quality )
  elseif ( line:match( "; Sparse Speed " ) ) then -- use the fastest one as the limit if we are doing the ceiling
		-- parse the setting
		maxSparseSpeed_MMpSEC = getNumericSettingValue( line )
		--print( "Max Speed " .. maxSparseSpeed_MMpSEC )

	elseif ( line:match( "; flow_max_mm3_per_s" ) ) then
		-- parse the setting
		maxFlowRate_MM3pSEC = getNumericSettingValue( line )
		--print( "Max Flow " .. maxFlowRate_MM3pSEC )
    
	elseif ( line:match( "; extrusion_width" ) ) then
		-- parse the setting
		extrusionWidth_MM = getNumericSettingValue( line )
		--print( "Extrusion Width " .. extrusionWidth_MM )

	elseif ( line:match( "; layer_thickness_mm" ) ) then
		-- parse the setting
		layerThickness_MM = getNumericSettingValue( line )
		print( "Layer Thickness " .. layerThickness_MM )
    
	elseif ( line:match( "; max_layer_thickness_mm" ) ) then
    -- THIS IS HOW WE TELL IF WE'RE USING KISS 1.6
    SLOWDOWN_FOR_FULL_FAN = 1.0           -- override the default so we return to the set time per layer

		-- parse the setting
		maxLayerThickness_MM = getNumericSettingValue( line )
    -- it's possible that it's not set i.e. 0
    if ( maxLayerThickness_MM < layerThickness_MM ) then
      maxLayerThickness_MM = layerThickness_MM
    end
		print( "Max Layer Thickness " .. maxLayerThickness_MM )

	elseif ( line:match( "; fiber_dia_mm" ) ) then
		-- parse the setting
		filamentDiameter_MM = getNumericSettingValue( line )
		--print( "Filament Diameter " .. filamentDiameter_MM )

	elseif ( line:match( "; min_layer_time_s" ) ) then
		kissMinLayerTime = getNumericSettingValue( line ) 
		--print( "Kisslicer min LayerTime " .. minLayerTime ) 
    ourMinLayerTime = kissMinLayerTime * SLOWDOWN_FOR_FULL_FAN * TIME_ESTIMATE_OPTIMISM_FACTOR
  --print( "ourMinLayerTime", ourMinLayerTime )
 	elseif ( line:match( "; fan_loops_percent" ) ) then
		kissFanLoopPercent = getNumericSettingValue( line ) * .01
		--print( "Kisslicer fan_loops_percent " .. kissFanLoopPercent )
    
	elseif ( line:match( "; fan_inside_percent" ) ) then
		kissFanInsidePercent = getNumericSettingValue( line ) *.01
		--print( "Kisslicer fan_inside_percent " .. kissFanInsidePercent )
    
	elseif ( line:match( "; fan_cool_percent" ) ) then
		kissFanCoolPercent = getNumericSettingValue( line ) * .01
		--print( "Kisslicer fan_cool_percent " .. kissFanCoolPercent )
 

	elseif ( line:match( "^M201" ) ) then     -- must begin the line, or it might be commented out

		--print "found max Acceleration settings"
		-- parse the setting

		local gtype, codeNum, axis1, axis1val, axis2, axis2val = parseGCodeSetting2Value( line )
		--print( "Acceleration: " .. gtype .. codeNum .. axis1 .. axis1val .. axis2 .. axis2val )
		--print( "Acceleration: " .. makeGCode2Value( gtype, codeNum, axis1, axis1val, axis2, axis2val ) )
		if ( type ) then
			foundAccelSpec = true
			if ( axis1 == 'X' and axis2 == 'Y' ) then
				if ( minAccelX == nil or axis1val < minAccelX ) then minAccelX = axis1val end
				if ( maxAccelX == nil or axis1val > maxAccelX ) then maxAccelX = axis1val end
				-- axis 2 is Y
				if ( maxAccelY == nil or axis2val > maxAccelY ) then maxAccelY = axis2val end
				if ( minAccelY == nil or axis2val < minAccelY ) then minAccelY = axis2val end
				doEcho = false
			end
		end
	elseif ( line:match( "^M205" ) ) then     -- must begin the line, or it might be commented out

		--print "found max Jerk settings"
		-- parse the setting

		local gtype, codeNum, axis1, axis1val = parseGCodeSetting1Value( line )
		--print( "Jerk: " .. makeGCode1Value( gtype, codeNum, axis1, axis1val ) )
		--print( "Jerk: " .. makeGCode1Value( gtype, codeNum, axis1, axis1val, axis2, axis2val ) )
		if ( type ) then
			foundJerkSpec = true
			if ( axis1 == 'X' ) then
				if ( maxJerkXY == nil or axis1val > maxJerkXY ) then maxJerkXY = axis1val end
				if ( minJerkXY == nil or axis1val < minJerkXY ) then minJerkXY = axis1val end
				doEcho = false
			end
		end

	elseif ( line:match( "Main%sG%-code" ) ) then

		AfterInit()
		--print "found Main G-Code"
		--whether we have found the spec or not, write out a line that sets it, based on the quality setting
		writeAccelSettings( fout, 1.0 )  -- full speed settings to begin with
		inMainGCode = true

		--writeAccelSettings( pctFast )
    
    --local filamentArea_MM2 = filamentDiameter_MM * filamentDiameter_MM * math.pi / 4
    --local maxExtruderAdvanceRate_MMpSEC = maxFlowRate_MM3pSEC / filamentArea_MM2
    -- TODO: make this adaptive as layer thickness changes, not just the maximum (to speed up thin regions)
    local drawPathArea_MM2 = math.max( layerThickness_MM, maxLayerThickness_MM ) * extrusionWidth_MM
    local maxDrawSpeed_MMpSEC = maxFlowRate_MM3pSEC / drawPathArea_MM2
    maxDrawSpeedForFlowRateLimit_MMpMIN = math.min( maxDrawSpeed_MMpSEC, maxSparseSpeed_MMpSEC ) * 60

    --print( "filamentArea_MM2", filamentArea_MM2 );
    --print( "maxExtruderAdvanceRate_MMpSEC", maxExtruderAdvanceRate_MMpSEC );
    --print( "drawPathArea_MM2", drawPathArea_MM2 );
    --print( "maxDrawSpeed_MMpSEC", maxDrawSpeed_MMpSEC );
    --print( "maxDrawSpeedForFlowRateLimit_MMpMIN", maxDrawSpeedForFlowRateLimit_MMpMIN );


--[[	elseif ( line:match( SPARSEFILL_PATTERN ) ) then

		--print "found sparse infill"
		fout:write( line .. "\n" )
		-- at very high speeds, the sparse infill parallel lines that are drawn to bridge
		-- above the stcaked infill do not stick well at all.
		-- We may need to slow them down, but first try just using less acceleration.
		local MAX_SPARSE_PctFast = .75  -- just clamp it a bit for the first sparse infill above stacked infill
		--local MAX_SPARSE_PctFast = .1  -- clamp it for the first sparse infill above stacked infill
		local pctFast = 1.0 - quality
		pctFast = math.min( pctFast, MAX_SPARSE_PctFast );
		writeAccelSettings( pctFast )  -- modify the accelerations
]]--

	end

	if ( doEcho == true ) then
		fout:write( line, "\n" )  -- echo it out
	end

end

-- done writing
fin:close();
fout:close();

-- this is a bit redundant. Try it out.
local leaveOriginal = not replaceOriginal -- if we are not replacing it (e.g. testing) then don't remove it either

if ( replaceOriginal == true ) then
	-- rename the original file
	local newOriginalName = inFileName .. "x"     -- just change the extension
	if ( leaveOriginal == true ) then
		os.remove( newOriginalName );                  -- get rid of any older original file if there is one
		os.rename( inFileName, newOriginalName );  -- rename the original to the new name
	else
		os.remove( newOriginalName );                  -- get rid of any older original file if there is one
		os.remove( inFileName );                  -- get rid of the one we modified so the directory doesn't get cluttered
	end
	os.rename( outFileName, inFileName );        -- and replace original with processed one
end

print( "Fastest possible Print Time: " .. math.floor(totalPrintTime_SEC/60) .. " min" )
print "done"

if ( replaceOriginal ~= true ) then
	print(string.format("elapsed time: %.2f\n", os.clock() - x))
	--local answer
	--if ( arg[1] ~= null ) then
	--	answer = io.read()
	--end
end

-- wait for a key
if ( debugging ) then
  answer = io.read()
end




